@@ -1,8 +1,8 @@
 #!/usr/bin/perl -w
 
-# $Id: oinkmaster.pl,v 1.378 2005/04/12 17:50:36 andreas_o Exp $ #
+# $Id: oinkmaster.pl,v 1.406 2006/02/10 13:02:44 andreas_o Exp $ #
 
-# Copyright (c) 2001-2005 Andreas stling <andreaso@it.su.se>
+# Copyright (c) 2001-2006 Andreas stling <andreaso@it.su.se>
 # All rights reserved.
 #
 #  Redistribution and use in source and binary forms, with or
@@ -65,9 +65,10 @@ sub print_summary_change($ $);
 sub make_backup($ $);
 sub get_changes($ $ $);
 sub update_rules($ @);
+sub copy_rules($ $);
 sub is_in_path($);
 sub get_next_entry($ $ $ $ $ $);
-sub get_new_vars($ $ $);
+sub get_new_vars($ $ $ $);
 sub add_new_vars($ $);
 sub write_new_vars($ $);
 sub msdos_to_cygwin_path($);
@@ -75,15 +76,16 @@ sub parse_mod_expr($ $ $ $);
 sub untaint_path($);
 sub approve_changes();
 sub parse_singleline_rule($ $ $);
+sub join_multilines($);
 sub minimize_diff($ $);
 sub catch_sigint();
 sub clean_exit($);
 
 
-my $VERSION            = 'Oinkmaster v1.2, Copyright (C) 2001-2005 Andreas stling <andreaso@it.su.se>';
+my $VERSION            = 'Oinkmaster v2.0, Copyright (C) 2001-2006 '.
+                         'Andreas stling <andreaso@it.su.se>';
 my $OUTFILE            = 'snortrules.tar.gz';
 my $RULES_DIR          = 'rules';
-my $DIST_SNORT_CONF    = 'snort.conf';
 
 my $PRINT_NEW          = 1;
 my $PRINT_OLD          = 2;
@@ -123,7 +125,7 @@ my $SINGLELINE_RULE_REGEXP = '^\s*#*\s*(?:%ACTIONS%)'.
                              '\s.+;\s*\)\s*$'; # ';
 
 # Match var line where var name goes into $1.
-my $VAR_REGEXP = '^\s*var\s+(\S+)\s+\S+';
+my $VAR_REGEXP = '^\s*var\s+(\S+)\s+(\S+)';
 
 # Allowed characters in misc paths/filenames, including the ones in the tarball.
 my $OK_PATH_CHARS = 'a-zA-Z\d\ _\(\)\[\]\.\-+:\\\/~@,=';
@@ -134,6 +136,10 @@ my @DEFAULT_CONFIG_FILES = qw(
     /usr/local/etc/oinkmaster.conf
 );
 
+my @DEFAULT_DIST_VAR_FILES = qw(
+  snort.conf
+);
+
 my (%loaded, $tmpdir);
 
 
@@ -165,11 +171,18 @@ if ($#{$config{config_files}} == -1) {
     }
 }
 
+# If no dist var file was specified on command line, set to default file(s).
+if ($#{$config{dist_var_files}} == -1) {
+    foreach my $var_file (@DEFAULT_DIST_VAR_FILES) {
+        push(@{${config{dist_var_files}}}, $var_file);
+    }
+}
+
 # If config is still not defined, we can't continue.
 if ($#{$config{config_files}} == -1) {
     clean_exit("configuration file not found in default locations\n".
                "(@DEFAULT_CONFIG_FILES)\n".
-               "Put it there or use the \"-C\" argument.");
+               "Put it there or use the \"-C <file>\" argument.");
 }
 
 read_config($_, \%config) for @{$config{config_files}};
@@ -220,8 +233,14 @@ foreach my $url (@{$config{url}}) {
     my $url_tmpdir = tempdir("url.XXXXXXXXXX", DIR => $tmpdir)
       or clean_exit("could not create temporary directory in $tmpdir: $!");
     push(@url_tmpdirs, "$url_tmpdir/$RULES_DIR");
-    download_file($url, "$url_tmpdir/$OUTFILE");
-    unpack_rules_archive("$url", "$url_tmpdir/$OUTFILE", $RULES_DIR);
+    if ($url =~ /^dir:\/\/(.+)/) {
+        mkdir("$url_tmpdir/$RULES_DIR")
+          or clean_exit("Could not create $url_tmpdir/$RULES_DIR");
+        copy_rules($1, "$url_tmpdir/$RULES_DIR");
+    } else {
+        download_file($url, "$url_tmpdir/$OUTFILE");
+        unpack_rules_archive("$url", "$url_tmpdir/$OUTFILE", $RULES_DIR);
+    }
 }
 
 # Copy all rules files from the tmp dirs into $RULES_DIR in the tmp directory.
@@ -235,7 +254,7 @@ clean_exit("not enough rules files in downloaded rules archive(s).\n".
            "Number of rules files is $num_files but minimum is set to $config{min_files}.")
   if ($num_files < $config{min_files});
 
-# For possible localized rules.
+# This is to read in possible 'localsid' rules.
 my %rh_tmp = setup_rules_hash(\%new_files, $config{output_dir});
 
 # Disable/modify/clean downloaded rules.
@@ -251,14 +270,14 @@ clean_exit("not enough rules in downloaded archive(s).\n".
            "Number of rules is $num_rules but minimum is set to $config{min_rules}.")
   if ($num_rules < $config{min_rules});
 
-# Setup a hash containing the content of all rules files.
+# Setup a hash containing the content of all processed rules files.
 my %rh = setup_rules_hash(\%new_files, $config{output_dir});
 
 # Compare the new rules to the old ones.
 my %changes = get_changes(\%rh, \%new_files, $RULES_DIR);
 
 # Check for variables that exist in dist snort.conf(s) but not in local snort.conf.
-get_new_vars(\%changes, $config{varfile}, \@url_tmpdirs)
+get_new_vars(\%changes, \@{$config{dist_var_files}}, $config{varfile}, \@url_tmpdirs)
   if ($config{update_vars});
 
 
@@ -302,7 +321,7 @@ if ($something_changed) {
       if ($config{make_backup} && !$config{quiet});
 }
 
-print "\nNote: Oinkmaster is running in careful mode - not updating anything.\n"
+print "\nOinkmaster is running in careful mode - not updating anything.\n"
   if ($something_changed && $config{careful});
 
 print_changes(\%changes, \%rh)
@@ -334,20 +353,23 @@ This should be the directory where you store your Snort rules.
 Options:
 -b <dir>  Backup your old rules into <dir> before overwriting them
 -c        Careful mode (dry run) - check for changes but do not update anything
--C <cfg>  Use this configuration file instead of the default
+-C <file> Use this configuration file instead of the default
           May be specified multiple times to load multiple files
 -e        Enable all rules that are disabled by default
 -h        Show this usage information
 -i        Interactive mode - you will be asked to approve the changes (if any)
 -m        Minimize diff when printing result by removing common parts in rules
 -q        Quiet mode - no output unless changes were found
--Q        super-quiet mode - like -q but even more quiet
+-Q        Super-quiet mode - like -q but even more quiet
 -r        Check for rules files that exist in the output directory
           but not in the downloaded rules archive
 -s        Leave out details in rules results, just print SID, msg and filename
+-S <file> Look for new variables in this file in the downloaded archive instead
+          of the default (@DEFAULT_DIST_VAR_FILES). Used in conjunction with -U.
+          May be specified multiple times to search multiple files.
 -T        Config test - just check configuration file(s) for errors/warnings
 -u <url>  Download from this URL instead of URL(s) in the configuration file
-          (http://, https://, ftp://, file:// or scp:// ... .tar.gz|.gz)
+          (http|https|ftp|file|scp:// ... .tar.gz|.gz, or dir://<dir>)
           May be specified multiple times to grab multiple rules archives
 -U <file> Merge new variables from downloaded snort.conf(s) into <file>
 -v        Verbose mode (debug)
@@ -377,8 +399,9 @@ sub parse_cmdline($)
         "o=s" => \$$cfg_ref{output_dir},
         "q"   => \$$cfg_ref{quiet},
         "Q"   => \$$cfg_ref{super_quiet},
-        "r"   => \$$cfg_ref{check_removed},
+         "r"   => \$$cfg_ref{check_removed},
         "s"   => \$$cfg_ref{summary_output},
+        "S=s" => \@{$$cfg_ref{dist_var_files}},
         "T"   => \$$cfg_ref{config_test_mode},
         "u=s" => \@{$$cfg_ref{url}},
         "U=s" => \$$cfg_ref{varfile},
@@ -389,6 +412,7 @@ sub parse_cmdline($)
                  }
     );
 
+
     show_usage unless ($cmdline_ok && $#ARGV == -1);
 
     $$cfg_ref{quiet}       = 1 if ($$cfg_ref{super_quiet});
@@ -399,6 +423,12 @@ sub parse_cmdline($)
         $$cfg_ref{make_backup} = 1;
     }
 
+  # Cannot specify dist var files without specifying var target file.
+    if (@{$$cfg_ref{dist_var_files}} && !$$cfg_ref{update_vars}) {
+        clean_exit("You can not specify distribution variable file(s) without ".
+                   "also specifying local file to merge into");
+    }
+
   # -o <dir> is the only required option in normal usage.
     if ($$cfg_ref{output_dir}) {
         $$cfg_ref{output_dir} = File::Spec->canonpath($$cfg_ref{output_dir});
@@ -427,6 +457,9 @@ sub read_config($ $)
     clean_exit("configuration file \"$config_file\" does not exist.\n")
       unless (-e "$config_file");
 
+    clean_exit("\"$config_file\" is not a file.\n")
+      unless (-f "$config_file");
+
     print STDERR "Loading $config_file\n"
       unless ($config{quiet});
 
@@ -449,21 +482,29 @@ sub read_config($ $)
     LINE:while ($_ = shift(@conf)) {
         $linenum++;
 
-      # Remove leading whitespaces and comment-only lines unless
-      # we're in the middle of reading a multi-line rule.
         unless ($multi) {
             s/^\s*//;
             s/^#.*//;
         }
 
-      # Multi-line start/continuation?
+      # Multi-line start/continuation.
         if (/\\\s*\n$/) {
             s/\\\s*\n$//;
+            s/^\s*#.*//;
+
+          # Be strict about removing #comments in modifysid/define_template statements, as
+          # they may contain other '#' chars.
+            if (defined($multi) && ($multi =~ /^modifysid/i || $multi =~ /^define_template/i)) {
+                s/#.*// if (/^\s*\d+[,\s\d]+#/);
+            } else {
+                s/\s*\#.*// unless (/^modifysid/i || /^define_template/i);
+            }
+
             $multi .= $_;
             next LINE;
         }
 
-      # Last line of multi-line directive?
+      # Last line of multi-line directive.
         if (defined($multi)) {
             $multi .= $_;
             $_ = $multi;
@@ -539,7 +580,8 @@ sub read_config($ $)
   	        if ($sid =~ /^\d+$/) {
                     $$cfg_ref{sid_disable_list}{$sid}++;
 	        } else {
-                    warn("WARNING: line $linenum in $config_file is invalid, ignoring\n");
+                    warn("WARNING: line $linenum in $config_file: ".
+                         "\"$sid\" is not a valid SID, ignoring\n");
 	        }
 	    }
 
@@ -550,7 +592,8 @@ sub read_config($ $)
   	        if ($sid =~ /^\d+$/) {
                     $$cfg_ref{sid_local_list}{$sid}++;
 	        } else {
-                    warn("WARNING: line $linenum in $config_file is invalid, ignoring\n");
+                    warn("WARNING: line $linenum in $config_file: ".
+                         "\"$sid\" is not a valid SID, ignoring\n");
 	        }
 	    }
 
@@ -561,7 +604,8 @@ sub read_config($ $)
   	        if ($sid =~ /^\d+$/) {
                     $$cfg_ref{sid_enable_list}{$sid}++;
 	        } else {
-                    warn("WARNING: line $linenum in $config_file is invalid, ignoring\n");
+                    warn("WARNING: line $linenum in $config_file: ".
+                         "\"$sid\" is not a valid SID, ignoring\n");
 	        }
 	    }
 
@@ -611,6 +655,9 @@ sub read_config($ $)
         } elsif (/^use_path_checks\s*=\s*([01])/i) {
             $$cfg_ref{use_path_checks} = $1;
 
+        } elsif (/^user_agent\s*=\s*(.+)/i) {
+            $$cfg_ref{user_agent} = $1;
+
         } elsif (/^include\s+(\S+.*)/i) {
              my $include = $1;
              read_config($include, $cfg_ref);
@@ -639,7 +686,7 @@ sub sanity_check()
 
   # Make sure all required variables are defined in the config file.
     foreach my $param (@req_params) {
-        clean_exit("the required parameter \"$param\" is not defined in configuration file.")
+        clean_exit("the required parameter \"$param\" is not defined in the configuration file.")
           unless (exists($config{$param}));
     }
 
@@ -683,8 +730,24 @@ sub sanity_check()
         clean_exit("variable file \"$config{varfile}\" does not exist.")
           unless (-e "$config{varfile}");
 
+        clean_exit("variable file \"$config{varfile}\" is not a file.")
+          unless (-f "$config{varfile}");
+
         clean_exit("variable file \"$config{varfile}\" is not writable by you.")
           if (!$config{careful} && !-w "$config{varfile}");
+
+      # Make sure dist var files don't contain [back]slashes
+      # (probably means user confused it with local var file).
+        my %dist_var_files;
+        foreach my $dist_var_file (@{${config{dist_var_files}}}) {
+            clean_exit("variable file \"$dist_var_file\" specified multiple times")
+              if (exists($dist_var_files{$dist_var_file}));
+            $dist_var_files{$dist_var_file} = 1;
+            clean_exit("variable file \"$dist_var_file\" contains slashes or backslashes ".
+                       "but it must be specified as a filename (without path) ".
+                       "that exists in the downloaded rules, e.g. \"snort.conf\"")
+              if ($dist_var_file =~ /\// || $dist_var_file =~ /\\/);
+        }
     }
 
   # Make sure all required binaries can be found, unless
@@ -707,8 +770,22 @@ sub sanity_check()
     $#{$config{url}} = -1;
     foreach my $url (@urls) {
         clean_exit("incorrect URL: \"$url\"")
-          unless ($url =~ /^((?:https*|ftp|file|scp):\/\/.+\.(?:tar\.gz|tgz))$/);
-        push(@{$config{url}}, $1);
+          unless ($url =~ /^((?:https*|ftp|file|scp):\/\/.+\.(?:tar\.gz|tgz))$/
+            || $url =~ /^(dir:\/\/.+)/);
+        my $ok_url = $1;
+
+        if ($ok_url =~ /^dir:\/\/(.+)/) {
+            my $dir = untaint_path($1);
+            clean_exit("\"$dir\" does not exist or is not a directory")
+              unless (-d $dir);
+
+          # Simple check if the output dir is specified as url (probably a mistake).
+            if (File::Spec->canonpath(File::Spec->rel2abs($dir))
+              eq File::Spec->canonpath(File::Spec->rel2abs($config{output_dir}))) {
+                clean_exit("Download directory can not be same as output directory");
+            }
+        }
+        push(@{$config{url}}, $ok_url);
     }
 
   # Wget must be found if url is http[s]:// or ftp://.
@@ -752,15 +829,15 @@ sub sanity_check()
             unless (-r "$config{output_dir}/$_") {
                 closedir(OUTDIR);
                 clean_exit("no read permission on \"$config{output_dir}/$_\"\n".
-                           "(Read permission is required on all rules files ".
-                           "inside the output directory.)\n")
+                           "Read permission is required on all rules files ".
+                           "inside the output directory.\n")
             }
 
             if (!$config{careful} && !-w "$config{output_dir}/$_") {
                 closedir(OUTDIR);
                 clean_exit("no write permission on \"$config{output_dir}/$_\"\n".
-                           "(Write permission is required on all rules files ".
-                           "inside the output directory.)\n")
+                           "Write permission is required on all rules files ".
+                           "inside the output directory.\n")
             }
 	}
     }
@@ -810,6 +887,13 @@ sub download_file($ $)
     $obfuscated_url = "$1:*password*\@$2"
       if ($obfuscated_url =~ /^(\S+:\/\/.+?):.+?@(.+)/);
 
+  # Ofbuscate oinkcode as well.
+    $obfuscated_url = "$1*oinkcode*$2"
+      if ($obfuscated_url =~ /^(\S+:\/\/.+\.cgi\/)[0-9a-z]{32,64}(\/.+)/i);
+
+    my @user_agent_opt;
+    @user_agent_opt = ("-U", $config{user_agent}) if (exists($config{user_agent}));
+
   # Use wget if URL starts with "http[s]" or "ftp" and we use external binaries.
     if ($config{use_external_bins} && $url =~ /^(?:https*|ftp)/) {
         print STDERR "Downloading file from $obfuscated_url... "
@@ -817,15 +901,25 @@ sub download_file($ $)
 
         if ($config{verbose}) {
             print STDERR "\n";
+            my @wget_cmd = ("wget", "-v", "-O", $localfile, $url, @user_agent_opt);
             clean_exit("could not download from $obfuscated_url")
-              if (system("wget", "-v", "-O", "$localfile", "$url"));
+              if (system(@wget_cmd));
+
         } else {
-            if (system("wget", "-v", "-o", "$log", "-O", "$localfile", "$url")) {
+            my @wget_cmd = ("wget", "-v", "-o", $log, "-O", $localfile, $url, @user_agent_opt);
+            if (system(@wget_cmd)) {
+                my $log_output;
                 open(LOG, "<", "$log")
                   or clean_exit("could not open $log for reading: $!");
-                my @log = <LOG>;
+                # Sanitize oinkcode in wget's log (password is automatically sanitized).
+                while (<LOG>) {
+                    $_ = "$1*oinkcode*$2"
+                      if (/(\S+:\/\/.+\.cgi\/)[0-9a-z]{32,64}(\/.+)/i);
+                    $log_output .= $_;
+                }
                 close(LOG);
-                clean_exit("could not download from $obfuscated_url. Output from wget follows:\n\n @log");
+                clean_exit("could not download from $obfuscated_url. ".
+                           "Output from wget follows:\n\n $log_output");
             }
             print STDERR "done.\n" unless ($config{quiet});
         }
@@ -835,7 +929,10 @@ sub download_file($ $)
         print STDERR "Downloading file from $obfuscated_url... "
           unless ($config{quiet});
 
-        my $ua = LWP::UserAgent->new();
+        my %lwp_opt;
+        $lwp_opt{agent} = $config{user_agent} if (exists($config{user_agent}));
+
+        my $ua = LWP::UserAgent->new(%lwp_opt);
         $ua->env_proxy;
 	my $request = HTTP::Request->new(GET => $url);
 	my $response = $ua->request($request, $localfile);
@@ -1006,7 +1103,8 @@ sub unpack_rules_archive($ $ $)
  # use_external_bins=0
     } else {
         $tar = Archive::Tar->new($archive, 1);
-        clean_exit("$url: could not read $archive\n")
+        clean_exit("$url: failed to read $archive (file transfer failed or ".
+                   "file in URL not in tar'ed gzip format?).")
           unless (defined($tar));
         @tar_content = $tar->list_files();
     }
@@ -1049,6 +1147,11 @@ sub unpack_rules_archive($ $ $)
 
             my $content = $tar->get_content($file);
 
+          # Symlinks in the archive will make get_content return undef.
+            clean_exit("could not get content from file \"$file\" in downloaded archive, ".
+                       "make sure it is a regular file\n")
+              unless (defined($content));
+
             open(RULEFILE, ">", "$file")
               or clean_exit("could not open \"$file\" for writing: $!\n");
             print RULEFILE $content;
@@ -1156,7 +1259,8 @@ sub process_rules($ $ $ $ $ $)
 
           # Is it a dup? If so, see if this seems to be more recent (higher rev).
             if (exists($sids{$sid})) {
-                warn("\nWARNING: duplicate SID in downloaded archive, SID=$sid\n")
+                warn("\nWARNING: duplicate SID in downloaded archive, SID=$sid, ".
+                     "only keeping rule with highest 'rev'\n")
                   unless($config{super_quiet});
 
                 my ($old_rev) = ($sids{$sid}{single} =~ /\brev\s*:\s*(\d+)\s*;/);
@@ -1295,9 +1399,8 @@ sub process_rules($ $ $ $ $ $)
         }
     }
 
-  # Print warnings on attempt at modifysid'ing non-existent stuff in all modes
-  # except super quiet, as they are usually more important.
-    unless ($config{super_quiet}) {
+  # Print warnings on attempt at modifysid'ing non-existent stuff, unless quiet mode.
+    unless ($config{quiet}) {
         my %new_files;
         foreach my $file (sort(keys(%$newfiles_ref))) {
             $new_files{basename($file)} = 1;
@@ -1359,10 +1462,13 @@ sub process_rule($ $ $ $ $ $ $ $)
   # modifysid that's no good since we don't know where in the
   # rule the trailing backslashes and newlines are going to be
   # and we don't want them to affect the regexp.
-    foreach my $mod_expr (@$modify_sid_ref) {
+    MOD_EXP:foreach my $mod_expr (@$modify_sid_ref) {
         my ($subst, $repl, $type, $arg) =
           ($mod_expr->[0], $mod_expr->[1], $mod_expr->[2], $mod_expr->[3]);
 
+        my $print_modify_warnings = 0;
+        $print_modify_warnings = 1 if (!$config{super_quiet} && $print_messages && $type eq "sid");
+
         if ($type eq "wildcard" || ($type eq "sid" && $sid eq $arg) ||
           ($type eq "file" && $filename eq $arg)) {
 
@@ -1372,6 +1478,17 @@ sub process_rule($ $ $ $ $ $ $ $)
                              "repl=$repl\nBefore: $single"
                   if ($print_messages && $config{verbose});
 
+
+              # If user specified a backreference but the regexp did not set $1 - don't modify rule.
+                if (!defined($1) && ($repl =~ /[^\\]\$\d+/ || $repl =~ /[^\\]\$\{\d+\}/
+                  || $repl =~ /^qq\/\$\d+/ || $repl =~ /^qq\/\$\{\d+\}/)) {
+                    warn("WARNING: SID $sid matches modifysid expression \"$subst\" but ".
+                         "backreference variable \$1 is undefined after match, ".
+                         "keeping original rule\n")
+                      if ($print_modify_warnings);
+                    next MOD_EXP;
+                }
+
               # Do the substitution on the single-line version and put it
               # back in $multi.
                 $single =~ s/$subst/$repl/eei;
@@ -1382,9 +1499,9 @@ sub process_rule($ $ $ $ $ $ $ $)
 
                 $$stats_ref{modified}++;
             } else {
-                if (!$config{super_quiet} && $print_messages && $type eq "sid") {
-                    print STDERR "WARNING: SID $sid does not match ".
-                                 "modifysid expression \"$subst\", skipping.\n";
+                if ($print_modify_warnings) {
+                    warn("WARNING: SID $sid does not match modifysid ".
+                         "expression \"$subst\", keeping original rule\n");
                 }
             }
         }
@@ -1756,8 +1873,6 @@ sub print_changes($ $)
           unless ($config{super_quiet} || $config{summary_output});
     }
 
-
-
   # Print list of possibly removed files if requested.
     if ($config{check_removed}) {
         if (keys(%{$$ch_ref{removed_files}})) {
@@ -1841,7 +1956,7 @@ sub print_summary_change($ $)
         my @rule = $sidmap{$sid}{rule};
         my $file = $sidmap{$sid}{file};
 	get_next_entry(\@rule, undef, undef, undef, \(my $msg), undef);
-        printf("%8d - %s ($file)\n", $sid, $msg);
+        printf("%8d - %s (%s)\n", $sid, $msg, $file);
     }
 
     print "\n";
@@ -1877,7 +1992,7 @@ sub get_changes($ $ $)
 
         while ($_ = readdir(OLDRULES)) {
             next if (/^\.\.?$/);
-            $changes{removed_files}{"$_"}++
+            $changes{removed_files}{"$_"} = 1
               if (/$config{update_files}/ && 
                 !exists($config{file_ignore_list}{$_}) &&
                 !-e "$tmpdir/$rules_dir/$_");
@@ -1971,7 +2086,7 @@ sub update_rules($ @)
     my $dst_dir        = shift;
     my @modified_files = @_;
 
-    print STDERR "Updating rules... "
+    print STDERR "Updating local rules files... "
       if (!$config{quiet} || $config{interactive});
 
     foreach my $file_w_path (@modified_files) {
@@ -1984,6 +2099,46 @@ sub update_rules($ @)
 }
 
 
+# Simply copy rules files from one dir to another.
+# Links are not allowed.
+sub copy_rules($ $)
+{
+    my $src_dir = shift;
+    my $dst_dir = shift;
+
+    print STDERR "Copying rules from $src_dir... "
+      if (!$config{quiet} || $config{interactive});
+
+    opendir(SRC_DIR, $src_dir)
+      or clean_exit("could not open directory $src_dir: $!");
+
+    my $num_files = 0;
+    while ($_ = readdir(SRC_DIR)) {
+        next if (/^\.\.?$/ || exists($config{file_ignore_list}{$_})
+          || !/$config{update_files}/);
+
+      my $src_file = untaint_path("$src_dir/$_");
+
+      # Make sure it's a regular file.
+        unless (-f "$src_file" && !-l "$src_file") {
+            closedir(SRC_DIR);
+            clean_exit("\"$src_file\" is not a regular file.")
+        }
+
+        unless (copy($src_file, $dst_dir)) {
+            closedir(SRC_DIR);
+            clean_exit("could not copy \"$src_file\" to \"$dst_dir\"/: $!");
+        }
+        $num_files++;
+    }
+
+    closedir(SRC_DIR);
+
+    print STDERR "$num_files files copied.\n"
+      if (!$config{quiet} || $config{interactive});
+}
+
+
 
 # Return true if file is in PATH and is executable.
 sub is_in_path($)
@@ -2151,81 +2306,91 @@ sub get_next_entry($ $ $ $ $ $)
 
 
 
-# Look for variables that exist in dist snort.conf but not in local snort.conf.
-sub get_new_vars($ $ $)
+# Look for variables that exist in dist var files but not in local var file.
+sub get_new_vars($ $ $ $)
 {
-    my $ch_ref          = shift;
-    my $local_conf      = shift;
-    my $url_tmpdirs_ref = shift;
-
-    my @new_vars;
-    my %old_vars;
+    my $ch_ref             = shift;
+    my $dist_var_files_ref = shift;
+    my $local_var_file     = shift;
+    my $url_tmpdirs_ref    = shift;
 
+    my %new_vars;
+    my (%old_vars, %dist_var_files, %found_dist_var_files);
     my $confs_found = 0;
 
+
+  # Warn in case we can't find a specified dist file.
     foreach my $dir (@$url_tmpdirs_ref) {
-        if (-e "$dir/$DIST_SNORT_CONF") {
-            $confs_found++;
+        foreach my $dist_var_file (@$dist_var_files_ref) {
+            if (-e "$dir/$dist_var_file") {
+                $found_dist_var_files{$dist_var_file} = 1;
+                $confs_found++;
+            }
+        }
+    }
+
+    foreach my $dist_var_file (@$dist_var_files_ref) {
+        unless (exists($found_dist_var_files{$dist_var_file})) {
+            warn("WARNING: did not find variable file \"$dist_var_file\" in ".
+                 "downloaded archive(s)\n")
+              unless($config{quiet});
         }
     }
 
-    my $base_conf = basename($DIST_SNORT_CONF);
     unless ($confs_found) {
         unless ($config{quiet}) {
-            warn("WARNING: no $base_conf found in downloaded archive(s), ".
+            warn("WARNING: no variable files found in downloaded archive(s), ".
                  "aborting check for new variables\n");
             return;
         }
     }
 
-    unless ($config{quiet}) {
-        print STDERR "Found $confs_found $DIST_SNORT_CONF file";
-        print STDERR "s" if ($confs_found > 1);
-        print STDERR ", checking for new variables... ";
-    }
+  # Read in variable names from old (target) var file.
+    open(LOCAL_VAR_FILE, "<", "$local_var_file")
+      or clean_exit("could not open $local_var_file for reading: $!");
 
-  # Read in variable names from old file.
-    open(LOCAL_CONF, "<", "$local_conf")
-      or clean_exit("could not open $local_conf for reading: $!");
+    my @local_var_conf = <LOCAL_VAR_FILE>;
 
-    my @local_conf = <LOCAL_CONF>;
-
-    foreach $_ (@local_conf) {
+    foreach $_ (join_multilines(\@local_var_conf)) {
         $old_vars{lc($1)}++ if (/$VAR_REGEXP/i);
     }
 
-    close(LOCAL_CONF);
+    close(LOCAL_VAR_FILE);
 
-
-  # Read in variables from new files.
+  # Read in variables from new file(s).
     foreach my $dir (@$url_tmpdirs_ref) {
-        my $conf = "$dir/$DIST_SNORT_CONF";
-        if (-e "$conf") {
+        foreach my $dist_var_file (@$dist_var_files_ref) {
+            my $conf = "$dir/$dist_var_file";
+            if (-e "$conf") {
+                my $num_new = 0;
+                print STDERR "Checking downloaded $dist_var_file for new variables... "
+                  unless ($config{quiet});
 
-            open(DIST_CONF, "<", "$conf")
-              or clean_exit("could not open $conf for reading: $!");
+                open(DIST_CONF, "<", "$conf")
+                  or clean_exit("could not open $conf for reading: $!");
+                my @dist_var_conf = <DIST_CONF>;
+                close(DIST_CONF);
 
-            while ($_ = <DIST_CONF>) {
-                s/^\s*//;
-
-                if (/$VAR_REGEXP/i && !exists($old_vars{lc($1)})) {
-                    if (/\\\n/) {
-                        warn("\nWARNING: can not handle variables split over multiple lines\n")
-                          unless ($config{quiet});
-                    } else {
-                        push(@new_vars, $_)
+                foreach $_ (join_multilines(\@dist_var_conf)) {
+                    if (/$VAR_REGEXP/i && !exists($old_vars{lc($1)})) {
+                        my ($varname, $varval) = (lc($1), $2);
+                        if (exists($new_vars{$varname})) {
+                            warn("\nWARNING: new variable \"$varname\" is defined multiple ".
+                                 "times in downloaded files\n");
+                        }
+                        s/^\s*//;
+                        push(@{$$ch_ref{new_vars}}, "$_\n");
+                        $new_vars{$varname} = $varval;
+                        $num_new++;
                     }
                 }
-            }
 
-            close(DIST_CONF);
+                close(DIST_CONF);
+                print STDERR "$num_new new found.\n"
+                  unless ($config{quiet});
+            }
         }
     }
-
-    @{$$ch_ref{new_vars}} = @new_vars;
-
-    print STDERR "done.\n"
-      unless ($config{quiet});
 }
 
 
@@ -2240,6 +2405,9 @@ sub add_new_vars($ $)
 
     return unless ($#{$changes{new_vars}} > -1);
 
+    print STDERR "Adding new variables to $varfile... "
+      unless ($config{quiet});
+
     open(OLD_LOCAL_CONF, "<", "$varfile")
       or clean_exit("could not open $varfile for reading: $!");
     my @old_content = <OLD_LOCAL_CONF>;
@@ -2251,7 +2419,7 @@ sub add_new_vars($ $)
     my @old_vars = grep(/$VAR_REGEXP/i, @old_content);
 
 
-  # If any vars exist in old file, find last one before inserting new ones.
+  # If any vars exist in old file, put new vars right after them.
     if ($#old_vars > -1) {
         while ($_ = shift(@old_content)) {
             print NEW_LOCAL_CONF $_;
@@ -2266,6 +2434,9 @@ sub add_new_vars($ $)
 
     clean_exit("could not copy $tmp_varfile to $varfile: $!")
       unless (copy("$tmp_varfile", "$varfile"));
+
+    print STDERR "done.\n"
+      unless ($config{quiet});
 }
 
 
@@ -2297,6 +2468,8 @@ sub parse_mod_expr($ $ $ $)
     my $subst        = shift;  # regexp to look for
     my $repl         = shift;  # regexp to replace it with
 
+    my @tmp_mod_list;
+
     $sid_arg_list =~ s/\s+$//;
 
     foreach my $sid_arg (split(/\s*,\s*/, $sid_arg_list)) {
@@ -2308,6 +2481,27 @@ sub parse_mod_expr($ $ $ $)
 
         return (0) unless ($type);
 
+      # Sanity check to make sure user escaped at least all the "$" in $subst.
+        if ($subst =~ /[^\\]\$./ || $subst =~ /^\$/) {
+            warn("WARNING: unescaped \$ in expression \"$subst\", all special ".
+                 "characters must be escaped\n");
+            return (0);
+        }
+
+      # Only allow backreference variables. The check should at least catch some user typos.
+        if (($repl =~ /[^\\]\$(\D.)/ && $1 !~ /{\d/) || $repl =~ /[^\\]\$$/
+          || ($repl =~ /^\$(\D.)/ && $1 !~ /{\d/)) {
+            warn("WARNING: illegal replacement expression \"$repl\": unescaped \$ ".
+                 "that isn't a backreference\n");
+            return (0);
+        }
+
+      # Don't permit unescaped @.
+        if ($repl =~ /[^\\]\@/ || $repl =~ /^\@/) {
+            warn("WARNING: illegal replacement expression \"$repl\": unescaped \@\n");
+            return (0);
+        }
+
       # Make sure the regexp is valid.
         my $repl_qq = "qq/$repl/";
         my $dummy   = "foo";
@@ -2316,12 +2510,19 @@ sub parse_mod_expr($ $ $ $)
             $dummy =~ s/$subst/$repl_qq/ee;
         };
 
+      # We should probably check for warnings as well as errors...
         if ($@) {
             warn("Invalid regexp: $@");
             return (0);
         }
 
-        push(@$mod_list_ref, [$subst, $repl_qq, $type, $sid_arg]);
+        push(@tmp_mod_list, [$subst, $repl_qq, $type, $sid_arg]);
+    }
+
+  # If we come this far, all sids and the regexp were parsed successfully, so
+  # append them to real mod list array.
+    foreach my $mod_entry (@tmp_mod_list) {
+        push(@$mod_list_ref, $mod_entry);
     }
 
     return (1);
@@ -2497,6 +2698,22 @@ sub parse_singleline_rule($ $ $)
 
 
 
+# Merge multiline directives in an array by simply removing traling backslashes.
+sub join_multilines($)
+{
+    my $multiline_conf_ref = shift;
+    my $joined_conf = "";
+
+    foreach $_ (@$multiline_conf_ref) {
+        s/\\\s*\n$//;
+        $joined_conf .= $_;
+    }
+
+    return (split/\n/, $joined_conf);
+}
+
+
+
 # Catch SIGINT.
 sub catch_sigint()
 {
